import configparser
from os import stat_result
from pathlib import Path
from typing import IO

import boto3
from boto3.s3.transfer import TransferConfig
from boto3_type_annotations.s3 import Client
from botocore.response import StreamingBody
from pydantic_core import ValidationError

from .. import pys3fuse_config_file
from ..core.config import settings
from ..core.log_config import logger
from ..core.schemas import FileAttrs, NodeSchema, S3ListObjectsResponse

session = boto3.Session(
    aws_access_key_id=settings.S3_ACCESS_KEY,
    aws_secret_access_key=settings.S3_SECRET_KEY,
)


config_parser = configparser.ConfigParser()
config_parser.read(pys3fuse_config_file)

mt = config_parser["s3_transfer_config"].getint("multipart_threshold")
xc = config_parser["s3_transfer_config"].getint("max_concurrency")
mc = config_parser["s3_transfer_config"].getint("multipart_chunksize")

transfer_config = TransferConfig(
    multipart_threshold=mt,
    max_concurrency=xc,
    multipart_chunksize=mc,
)


class S3Service:
    def __init__(self, client: Client, bucket: str):
        self.client = client
        self.bucket = bucket
        self.config = transfer_config

    def get_bucket_acl(self):
        return self.client.get_bucket_acl(Bucket=self.bucket)

    def list_objects(self) -> S3ListObjectsResponse:
        try:
            response = S3ListObjectsResponse(
                **self.client.list_objects_v2(Bucket=self.bucket)
            )
            return response
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise

    def get_object(self, key: str) -> bytes:
        try:
            response: StreamingBody = self.client.get_object(
                Bucket=self.bucket, Key=key
            )["Body"]
            fileobj = response.read()
            return fileobj
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise

    def head_object(self, key: str) -> NodeSchema:
        try:
            response = self.client.head_object(Bucket=self.bucket, Key=key)
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise
        try:
            file_attrs = FileAttrs(**response["Metadata"])
        except ValidationError:
            file_attrs = None

        return NodeSchema(
            key=key,
            last_modified=response["LastModified"],
            size=response["ContentLength"],
            content_type=response["ContentType"],
            file_attrs=file_attrs,
        )

    def put_object(self, file_path: Path, stat: stat_result, mime_type: str):
        key = self.get_key_from_path(file_path)
        file_attrs = FileAttrs.from_stat_result(stat)
        try:
            with open(file_path, "br") as f:
                self.client.put_object(
                    Bucket=self.bucket,
                    Key=key,
                    Body=f.read(),
                    ContentType=mime_type,
                    Metadata=file_attrs.model_dump(),
                )
        except FileNotFoundError:
            return

    def upload_fileobj(
        self, fileobj: IO, key: str, file_attrs: FileAttrs, mime_type: str
    ):
        try:
            self.client.upload_fileobj(
                Fileobj=fileobj,
                Bucket=self.bucket,
                Key=key,
                ExtraArgs={
                    "ContentType": mime_type,
                    "Metadata": file_attrs.model_dump(),
                },
                Config=self.config,
            )
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            return

    def delete_objects(self, *keys) -> None:
        if not keys:
            return
        delete = {"Objects": [{"Key": k} for k in keys]}
        try:
            self.client.delete_objects(Bucket=self.bucket, Delete=delete)
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise

    def copy_object(self, key: str, attrs: FileAttrs, mime_type: str):
        try:
            self.client.copy_object(
                Bucket=self.bucket,
                CopySource={"Bucket": self.bucket, "Key": key},
                Key=key,
                ContentType=mime_type,
                Metadata=attrs.model_dump(),
                MetadataDirective="REPLACE",
            )
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise

    def rename_object(self, old_key: str, new_key: str):
        try:
            self.client.copy_object(
                Bucket=self.bucket,
                CopySource={"Bucket": self.bucket, "Key": old_key},
                Key=new_key,
            )
            self.client.delete_object(Bucket=self.bucket, Key=old_key)
        except Exception as e:
            logger.exception(str(e))
            raise

    def download_fileobj(self, fileobj: IO, key: str):
        try:
            self.client.download_fileobj(
                Fileobj=fileobj,
                Bucket=self.bucket,
                Key=key,
                Config=self.config,
            )
        except Exception as e:
            logger.exception(str(e), exc_info=e)
            raise

    @staticmethod
    def get_key_from_path(file_path):
        _, *dirs, _ = file_path.parts
        key = f"{'/'.join(dirs)}/{file_path.name}"
        return key



import configparser
import errno
import os
import queue
import threading
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Callable, Iterator

from fuse import FuseOSError, Operations, fuse_get_context

from .. import pys3fuse_config_file
from ..core.log_config import logger
from ..core.schemas import FileAttrs
from ..core.sync import SyncResult, Syncer
from ..core.utils import full_path, log, on_error
from ..s3.client import S3Service


@dataclass
class Task:
    target: Callable[..., SyncResult]
    args: tuple
    caller: str

    def __call__(self) -> SyncResult:
        return self.target(*self.args)


class Filesystem(Operations):
    def __init__(
        self,
        *,
        root: str,
        bucket: str,
        offline: bool = False,
        service: S3Service | None = None,
        multi_user: bool = False,
        force: bool = False,
    ):
        self.root = Path(root)
        self.bucket = bucket
        self.offline = offline
        self.service = service
        self.multi_user = multi_user
        self.force = force
        self.config_parser = configparser.ConfigParser()
        self.config_parser.read(pys3fuse_config_file)

        self.syncer = Syncer(self.service, self.root, self.config_parser)

        self.queue = queue.Queue[Task]()

        def worker():
            while True:
                item = self.queue.get()
                logger.debug(
                    f"Syncing: <{item.target.__name__}::{item.args}>",
                    extra={"worker_name": item.caller},
                )
                sync_result = item()
                logger.debug(sync_result, extra={"worker_name": item.caller})
                self.queue.task_done()

        threading.Thread(target=worker, daemon=True, name="Daemon").start()

    @cached_property
    def should_sync_multi_user(self):
        return (not self.offline) and self.multi_user

    @cached_property
    def should_sync(self):
        return not self.offline

    def _full_path(self, partial) -> Path:
        return self.root / partial.lstrip("/")

    @full_path
    @log
    def access(self, path: Path, mode: int):
        if not os.access(path, mode):
            raise FuseOSError(errno.EACCES)

    @full_path
    @log
    def chmod(self, path: Path, mode: int) -> None:
        st = os.lstat(path)
        try:
            os.chmod(path, mode)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                attrs = FileAttrs.from_stat_result(st)
                attrs.st_mode = mode
                self.queue.put_nowait(
                    Task(self.syncer.update_attrs_on_s3, (path, attrs), "chmod")
                )

    @full_path
    @log
    def chown(self, path: Path, uid: int, gid: int) -> None:
        st = os.lstat(path)

        def change_xid(uid_, gid_) -> tuple[int, int]:
            uid_ = st.st_uid if uid_ == -1 else uid_
            gid_ = st.st_gid if gid_ == -1 else gid_
            return uid_, gid_

        try:
            os.chown(path, uid, gid)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                attrs = FileAttrs.from_stat_result(st)
                u, g = change_xid(uid, gid)
                attrs.st_uid = u
                attrs.st_gid = g
                self.queue.put_nowait(
                    Task(self.syncer.update_attrs_on_s3, (path, attrs), "chown")
                )

    @log
    def getattr(self, path: str, fh=None) -> dict:
        full_path_ = self._full_path(path)
        st = os.lstat(full_path_)

        attrs = dict(
            (key, getattr(st, key))
            for key in (
                "st_atime",
                "st_ctime",
                "st_gid",
                "st_mode",
                "st_mtime",
                "st_nlink",
                "st_size",
                "st_uid",
            )
        )
        return attrs

    @full_path
    @log
    def readdir(self, path: Path, fh) -> Iterator[str]:
        if self.should_sync_multi_user:
            logger.debug(f"Syncing: : <sync_local_with_s3::{(self.force,)}>")
            sync_result = self.syncer.sync_local(self.force)
            logger.debug(sync_result)

        dirents = [".", ".."]
        if os.path.isdir(path):
            dirents.extend(os.listdir(path))
        for r in dirents:
            yield r

    @full_path
    @log
    def readlink(self, path) -> str:
        try:
            pathname = os.readlink(path)
        except OSError as exc:
            on_error(exc)
        else:
            if pathname.startswith("/"):
                return os.path.relpath(pathname, self.root)
            else:
                return pathname

    @full_path
    @log
    def mknod(self, path: Path, mode: int, dev) -> None:
        try:
            os.mknod(path, mode, dev)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def rmdir(self, path) -> None:
        try:
            os.rmdir(path)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def mkdir(self, path: Path, mode: int) -> None:
        try:
            os.mkdir(path, mode)
        except FileExistsError as exc:
            on_error(exc)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def statfs(self, path: Path) -> dict:
        try:
            stv = os.statvfs(path)
        except OSError as exc:
            on_error(exc)
        else:
            return dict(
                (key, getattr(stv, key))
                for key in (
                    "f_bavail",
                    "f_bfree",
                    "f_blocks",
                    "f_bsize",
                    "f_favail",
                    "f_ffree",
                    "f_files",
                    "f_flag",
                    "f_frsize",
                    "f_namemax",
                )
            )

    @full_path
    @log
    def unlink(self, path: Path) -> None:
        try:
            os.unlink(path)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                self.queue.put_nowait(Task(self.syncer.delete_on_s3, (path,), "unlink"))

    @log
    def symlink(self, name, target, *args) -> None:
        path = self._full_path(name)
        uid, gid, pid = fuse_get_context()
        try:
            os.symlink(target, path)
            os.chown(path, uid, gid, follow_symlinks=False)
        except OSError as exc:
            on_error(exc)

    @log
    def rename(self, old: str, new: str) -> None:
        old_path = self._full_path(old)
        new_path = self._full_path(new)

        try:
            os.rename(old_path, new_path)
        except FileExistsError as exc:
            on_error(exc)
        except IsADirectoryError as exc:
            on_error(exc)
        except NotADirectoryError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                self.queue.put_nowait(
                    Task(self.syncer.rename_on_s3, (old_path, new_path), "rename")
                )

    @log
    def link(self, target, name) -> None:
        try:
            os.link(self._full_path(name), self._full_path(target))
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def utimens(self, path: Path, times=None) -> None:
        try:
            os.utime(path, times)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                attrs = FileAttrs.from_stat_result(os.lstat(path))
                self.queue.put_nowait(
                    Task(self.syncer.update_attrs_on_s3, (path, attrs), "utimens")
                )

    @full_path
    @log
    def open(self, path: Path, flags) -> int:
        try:
            return os.open(path, flags)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def create(self, path: Path, mode: int, fi=None) -> int:
        uid, gid, pid = fuse_get_context()
        try:
            fd = os.open(path, os.O_WRONLY | os.O_CREAT, mode)
            os.chown(path, uid, gid)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                self.queue.put_nowait(Task(self.syncer.create_on_s3, (path,), "create"))
            return fd

    @log
    def read(self, path: str, length, offset, fh) -> bytes:
        os.lseek(fh, offset, os.SEEK_SET)
        return os.read(fh, length)

    @log
    def write(self, path: str, buf, offset, fh) -> int:
        os.lseek(fh, offset, os.SEEK_SET)
        return os.write(fh, buf)

    @full_path
    @log
    def truncate(self, path: Path, length, fh: int = None) -> None:
        try:
            with open(path, "r+") as f:
                f.truncate(length)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                self.queue.put_nowait(
                    Task(self.syncer.update_on_s3, (path,), "truncate")
                )

    @log
    def flush(self, path: str, fh) -> None:
        try:
            os.fsync(fh)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def release(self, path: Path, fh: int) -> None:
        try:
            os.close(fh)
        except OSError as exc:
            on_error(exc)
        else:
            if self.should_sync:
                self.queue.put_nowait(
                    Task(self.syncer.update_on_s3, (path,), "release")
                )

    @full_path
    @log
    def fsync(self, path: str, fdatasync, fh: int) -> None:
        self.flush(path, fh)



from datetime import datetime
from os import stat_result
from pathlib import Path
from typing import Any, List, Literal, Self, TYPE_CHECKING, Union

from pydantic import BaseModel

if TYPE_CHECKING:
    pass


class FileAttrs(BaseModel):
    st_atime: float
    st_ctime: float
    st_gid: int | None = None
    st_mode: int | None = None
    st_mtime: float
    st_nlink: int = 1
    st_size: int
    st_uid: int | None = None

    def model_dump(self, for_fs: bool = False) -> dict[str, Any]:  # noqa
        if for_fs:
            return super().model_dump()

        d = super().model_dump(exclude_none=True)
        for k, v in d.items():
            d[k] = str(v)
        return d

    @classmethod
    def from_stat_result(cls, stat_res: stat_result) -> Self:
        return cls(
            st_atime=stat_res.st_atime,
            st_ctime=stat_res.st_ctime,
            st_mtime=stat_res.st_mtime,
            st_mode=stat_res.st_mode,
            st_gid=stat_res.st_gid,
            st_uid=stat_res.st_uid,
            st_nlink=stat_res.st_nlink,
            st_size=stat_res.st_size,
        )

    @classmethod
    def from_node_schema(cls, node: "NodeSchema") -> Self:
        return cls(
            st_atime=node.last_modified.timestamp(),
            st_ctime=node.last_modified.timestamp(),
            st_mtime=node.last_modified.timestamp(),
            st_gid=None,
            st_uid=None,
            st_mode=33060,  # 0o444
            st_nlink=1,
            st_size=node.size,
        )


class OwnerModel(BaseModel):
    DisplayName: str | None = None
    ID: str | None = None


class RestoreStatusModel(BaseModel):
    IsRestoreInProgress: bool
    RestoreExpiryDate: datetime | None = None


class CommonPrefixesModel(BaseModel):
    Prefix: str


class Content(BaseModel):
    Key: str
    LastModified: datetime
    ETag: str | None = None
    ChecksumAlgorithm: (
        list[
            Literal[
                "CRC32",
                "CRC32C",
                "SHA1",
                "SHA256",
                "CRC64NVME",
            ]
        ]
        | None
    ) = None
    ChecksumType: Literal["COMPOSITE", "FULL_OBJECT"] | None = None
    Size: int
    StorageClass: (
        Literal[
            "STANDARD",
            "REDUCED_REDUNDANCY",
            "GLACIER",
            "STANDARD_IA",
            "ONEZONE_IA",
            "INTELLIGENT_TIERING",
            "DEEP_ARCHIVE",
            "OUTPOSTS",
            "GLACIER_IR",
            "SNOW",
            "EXPRESS_ONEZONE",
            "FSX_OPENZFS",
        ]
        | None
    ) = None
    Owner: OwnerModel | None = None
    RestoreStatus: RestoreStatusModel | None = None


class S3ListObjectsResponse(BaseModel):
    ResponseMetadata: dict | None = None
    IsTruncated: bool
    Contents: list[Content] | None = None
    Name: str
    Prefix: str | None = None
    Delimiter: str | None = None
    MaxKeys: int
    CommonPrefixes: List[CommonPrefixesModel] | None = None
    EncodingType: Literal["url"] | None = None
    KeyCount: int
    ContinuationToken: str | None = None
    NextContinuationToken: str | None = None
    StartAfter: str | None = None
    RequestCharged: Literal["requester"] | None = None


class NodeSchema(BaseModel):
    key: str
    last_modified: datetime
    size: int
    abspath: Path | None = None
    content_type: str | None = None
    file_attrs: Union["FileAttrs", None] = None



import configparser
import os
from pathlib import Path

from boto3.s3.transfer import MB

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Self Dir ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
pys3fuse_dir = Path(os.getenv("HOME")) / ".pys3fuse"

if not pys3fuse_dir.exists():
    os.mkdir(pys3fuse_dir, mode=0o766)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Logs ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
pys3fuse_logs_dir = pys3fuse_dir / "logs"

if not pys3fuse_logs_dir.exists():
    os.mkdir(pys3fuse_logs_dir, 0o766)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Config ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
pys3fuse_config_file = pys3fuse_dir / "config.ini"

if not pys3fuse_config_file.exists():
    config_parser = configparser.ConfigParser()
    config_parser["log"] = {"level": "info"}
    config_parser["s3_transfer_config"] = {
        "multipart_threshold": 8 * MB,
        "max_concurrency": 10,
        "multipart_chunksize": 8 * MB,
    }
    config_parser["sync"] = {"dirty": "false"}

    with pys3fuse_config_file.open("w") as f:
        config_parser.write(f)  # noqa



import configparser
import logging
from argparse import Namespace
from pathlib import Path

import botocore.config
import botocore.errorfactory
from boto3_type_annotations.s3 import Client
from botocore.exceptions import ClientError, EndpointConnectionError
from fuse import FUSE

from .. import pys3fuse_config_file
from ..core.config import settings
from ..core.exceptions import PysClientError, S3ConnectionError
from ..core.log_config import logger
from ..core.sync import Syncer
from ..core.utils import write_config
from ..fs.filesystem import Filesystem
from ..s3.client import S3Service, session


class MountHandler:
    def __init__(self):
        self.config_parser = configparser.ConfigParser()
        self.config_parser.read(pys3fuse_config_file)

        # fmt: off
        logger.setLevel(
            logging.INFO
            if self.config_parser["log"]["level"] == "info"
            else logging.DEBUG
        )
        # fmt: on

        self.client: Client | None = None
        self.service: S3Service | None = None

    def __call__(self, args: Namespace) -> None:
        if args.debug:
            logger.setLevel(logging.DEBUG)

        self.args = args
        try:
            if args.offline is True:
                self.config_parser["sync"]["dirty"] = "true"
            else:
                self.check_s3()
                syncer = Syncer(
                    self.service,
                    Path(args.source),
                    self.config_parser,
                )
                syncer.sync_initial(force_delete=args.force)
            write_config(self.config_parser)
            self.mount()
        except S3ConnectionError:
            logger.error(
                f"Exiting... <S3: Bucket={args.bucket}> is not available. "
                f"If you are not connected to the internet pass `--offline` option."
            )
        except PysClientError as e:
            logger.error(str(e))
        except Exception as e:
            logger.exception(str(e), exc_info=e)
        finally:
            print()
            logger.info("Unmounting...")
            if self.client is not None:
                self.client.close()

    def check_s3(self):
        """Check the connection between FS and S3"""
        args = self.args
        logger.debug("Checking S3 Connection...")
        try:
            self.client = client = session.client(
                "s3",
                endpoint_url=settings.S3_API,
                config=botocore.config.Config(
                    tcp_keepalive=True,
                    max_pool_connections=10,
                ),
            )
            self.service = service = S3Service(client, args.bucket)
            service.get_bucket_acl()
        except EndpointConnectionError:
            raise S3ConnectionError() from None
        except ClientError as e:
            raise PysClientError(str(e)) from None
        except Exception:
            raise
        else:
            logger.info("S3 Connection is established.")

    def mount(self):
        """Mount the FS"""
        logger.info("Mounting...")
        args = self.args
        FUSE(
            Filesystem(
                root=args.source,
                bucket=args.bucket,
                offline=args.offline,
                service=self.service,
                multi_user=args.multi_user,
                force=args.force,
            ),
            args.mountpoint,
            foreground=True,
            allow_other=True,
        )



import configparser
from argparse import ArgumentParser, Namespace
from typing import assert_never

import rich
from boto3.s3.transfer import KB, MB
from rich.syntax import Syntax

from pys3fuse import pys3fuse_config_file


class ConfigHandler:
    def __init__(self, parser: ArgumentParser):
        self.parser = parser

    def __call__(self, args: Namespace) -> None:
        config_parser = configparser.ConfigParser()
        config_parser.read(pys3fuse_config_file)

        if args.show:
            with open(pys3fuse_config_file) as f:
                rich.print(Syntax(f.read(), "ini", line_numbers=True))
            exit(0)

        if (
            args.log_level is None
            and args.multipart_threshold is None
            and args.max_concurrency is None
            and args.multipart_chunksize is None
            and args.create_on_the_fly is None
        ):
            self.parser.print_help()
            exit(-1)

        if (v := args.log_level) is not None:
            config_parser["log"]["level"] = v

        if (v := args.multipart_threshold) is not None:
            num, suf = int("".join(filter(str.isdigit, v))), v[-2:]  # noqa
            match suf:
                case "KB":
                    v = num * KB
                case "MB":
                    v = num * MB
                case _:
                    assert_never(suf)

            config_parser["s3_transfer_config"]["multipart_threshold"] = str(v)

        if (v := args.max_concurrency) is not None:
            config_parser["s3_transfer_config"]["max_concurrency"] = v

        if (v := args.multipart_chunksize) is not None:
            num, suf = int("".join(filter(str.isdigit, v))), v[-2:]  # noqa
            match suf:
                case "KB":
                    v = num * KB
                case "MB":
                    v = num * MB
                case _:
                    assert_never(suf)
            config_parser["s3_transfer_config"]["multipart_chunksize"] = str(v)

        with open(pys3fuse_config_file, "w") as f:
            config_parser.write(f)  # noqa



import logging
import queue
from datetime import datetime
from logging import LogRecord
from logging.config import dictConfig
from logging.handlers import QueueHandler, QueueListener, RotatingFileHandler
from typing import cast

import boto3
import botocore
import rich.logging
import structlog
import urllib3
from rich.console import Console, ConsoleRenderable, RenderableType
from rich.containers import Renderables
from rich.table import Table
from rich.text import Text, TextType
from rich.traceback import Traceback

from .. import pys3fuse_dir

cons = Console(force_terminal=True)


class DefaultLogRender:
    def __init__(self, time_format="[%x %X]") -> None:
        self.time_format = time_format
        self.omit_repeated_times = False
        self.level_width = None
        self.last_func_name = None

    def __call__(
        self,
        console: Console,
        renderables: list[ConsoleRenderable],
        log_time: datetime,
        level: TextType,
        record: LogRecord,
    ) -> Table:
        output = Table.grid(padding=(0, 1))
        output.expand = True
        output.add_column(style="log.time")
        output.add_column(style="log.level")
        output.add_column()
        output.add_column(ratio=1, style="log.message", overflow="fold")

        row: list[RenderableType] = []

        log_time = log_time or console.get_datetime()
        log_time_display = Text(log_time.strftime(self.time_format))
        row.append(log_time_display)
        row.append(level)

        func_name = getattr(record, "func_name", None)
        if func_name is None:
            func_name = record.funcName
            func_name_text = Text()
            if func_name == "__call__":
                func_name_text.append("<main>", style="bright_yellow bold")
            elif (wn := getattr(record, "worker_name", None)) is not None:
                func_name_text.append(f"<{wn}>", style="bright_yellow bold")
            else:
                func_name_text.append(f"<{func_name}>", style="bright_yellow bold")
        else:
            func_name_text = Text()
            func_name_text.append(f"[{func_name:<8}]", style="magenta1 bold")

        row.append(func_name_text)
        row.append(Renderables(renderables))
        output.add_row(*row)
        return output


class PyS3FUSERichHandler(rich.logging.RichHandler):
    def __init__(self):
        super().__init__(
            console=cons,
            markup=True,
            rich_tracebacks=True,
            tracebacks_show_locals=False,
            tracebacks_suppress=[boto3, botocore, urllib3],
            tracebacks_max_frames=1,
        )
        self._log_render = DefaultLogRender("%Y-%m-%d %H:%M:%S")

    def render(
        self,
        *,
        record: logging.LogRecord,
        traceback: Traceback | None,
        message_renderable: "ConsoleRenderable",
    ) -> "ConsoleRenderable":
        level = self.get_level_text(record)
        log_time = datetime.fromtimestamp(record.created)

        log_renderable = self._log_render(
            self.console,
            [message_renderable] if not traceback else [message_renderable, traceback],
            log_time,
            level,
            record,
        )
        return log_renderable


class PyS3FUSEQueueListener(QueueListener):
    def handle(self, record: LogRecord):
        record = self.prepare(record)
        handler = logging.getHandlerByName("console_handler")

        if not self.respect_handler_level:
            process = True
        else:
            process = record.levelno >= handler.level
        if process:
            handler.handle(record)


class PyS3FUSEQueueHandler(QueueHandler):
    listener: PyS3FUSEQueueListener

    def prepare(self, record: LogRecord) -> LogRecord:
        return record


class PyS3FUSEStructLogProcessor:
    def __call__(self, _, __, event_dict: dict):
        log_dict = {
            "timestamp": event_dict["timestamp"],
            "logger": event_dict["logger"],
            "level": event_dict["level"],
            "message": event_dict["message"],
        }
        if (exception := event_dict.get("exception", None)) is not None:
            log_dict["exception"] = exception
        return log_dict


dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "loggers": {
            "pys3fuse": {
                "level": logging.DEBUG,
                "handlers": ["console_handler", "file_queue_handler"],
            },
            "fuse": {
                "level": logging.CRITICAL,
                "handlers": ["console_handler", "file_queue_handler"],
            },
            "boto3": {
                "level": logging.ERROR,
                "handlers": ["console_handler", "file_queue_handler"],
            },
        },
        "handlers": {
            "console_handler": {"class": PyS3FUSERichHandler},
            "file_handler": {
                "class": RotatingFileHandler,
                "formatter": "file_formatter",
                "filename": pys3fuse_dir / "logs" / "pys3fuse.log",
                "maxBytes": 10 * 1024 * 1024,  # 10 MB
                "backupCount": 10,
                "encoding": "utf8",
            },
            "file_queue_handler": {
                "class": PyS3FUSEQueueHandler,
                "queue": queue.Queue(),
                "listener": QueueListener,
                "handlers": ["file_handler"],
            },
        },
        "formatters": {
            "file_formatter": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processors": [
                    structlog.stdlib.add_log_level,
                    structlog.stdlib.add_logger_name,
                    structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
                    structlog.processors.ExceptionRenderer(
                        structlog.processors.ExceptionDictTransformer()
                    ),
                    structlog.processors.EventRenamer("message"),
                    PyS3FUSEStructLogProcessor(),
                    structlog.processors.LogfmtRenderer(),
                ],
            },
        },
    }
)

logger = logging.getLogger("pys3fuse")
file_queue_listener: QueueListener = cast(  # noqa
    QueueHandler, logging.getHandlerByName("file_queue_handler")
).listener



import errno
import inspect
import subprocess
from configparser import ConfigParser
from functools import singledispatch, wraps
from pathlib import Path
from typing import Never, ParamSpec

from fuse import FuseOSError

from .log_config import logger
from .. import pys3fuse_config_file


@singledispatch
def on_error(exc: Exception, *, errno_: int = 0) -> Never:
    """Log Exception and raise `FuseOSError`"""
    raise FuseOSError(errno_)


@on_error.register
def on_os_error(exc: OSError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_key_error(exc: KeyError) -> Never:
    raise FuseOSError(errno.ENOENT)


@on_error.register
def on_is_a_dir_error(exc: IsADirectoryError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_not_a_dir_error(exc: NotADirectoryError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_file_not_found_error(exc: FileNotFoundError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_file_exists(exc: FileExistsError) -> Never:
    raise FuseOSError(exc.errno)


def log(func):
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs):
        func_sig = inspect.signature(func)
        params = dict(zip(func_sig.parameters, args))
        del params["self"]

        log_msg = ", ".join(f"{k}={v}" for k, v in params.items())
        logger.debug(log_msg, extra={"func_name": func.__name__})
        return func(*args, **kwargs)

    return inner


P = ParamSpec("P")


def full_path(func):
    def inner(*args: P.args, **kwargs: P.kwargs):
        func_sig = inspect.signature(func)
        params = dict(zip(func_sig.parameters, args))

        self = params["self"]
        params["path"] = self.root / params["path"].lstrip("/")

        args = tuple(params.values())
        return func(*args, **kwargs)

    return inner


def write_config(config_parser: ConfigParser) -> None:
    """Write the mutated cofig to the config file"""
    with open(pys3fuse_config_file, "w") as f:
        config_parser.write(f)  # noqa


def get_mime_type(file_path: Path) -> str:
    return subprocess.getoutput(f"file --mime-type -b {str(file_path.absolute())}")



import os
import threading
from concurrent.futures import ThreadPoolExecutor
from configparser import ConfigParser
from pathlib import Path
from typing import NamedTuple, Sized

from rich.prompt import Prompt

from .schemas import FileAttrs, NodeSchema
from ..core.log_config import logger
from ..core.utils import get_mime_type
from ..s3.client import S3Service


class SyncResult(NamedTuple):
    created_files: list[str] | set[str] = []
    updated_files: list[str] | set[str] = []
    deleted_files: list[str] | set[str] = []

    @property
    def created(self) -> int:
        return len(self.created_files)

    @property
    def updated(self):
        return len(self.updated_files)

    @property
    def deleted(self) -> int:
        return len(self.deleted_files)

    @property
    def synced(self) -> bool:
        return bool(self.created or self.updated or self.deleted)


class Syncer:
    def __init__(
        self,
        service: S3Service,
        root: Path,
        config_parser: ConfigParser,
    ):
        self.service = service
        self.root = root
        self.config_parser = config_parser
        self.lock = threading.Lock()

    def sync_initial(self, force_delete: bool = False):
        """Sync local or S3 respecting the dirtiness"""
        logger.debug("Getting nodes from S3...")

        s3_nodes = self._get_node_schema_from_bucket()
        s3_keys = {node.abspath for node in s3_nodes}
        source_keys = self._get_all_files(self.root)

        if self.config_parser["sync"]["dirty"] == "true":
            sync_result = self._sync_s3_with_local_objects(
                s3_nodes, s3_keys, source_keys, force_delete=force_delete
            )
            if sync_result:
                logger.debug(
                    "Local filesystem is dirty. We have to sync S3 with here so"
                    " that local changes would appear on S3.",
                )
                logger.info(f"Syncing S3 with {self.root.name}...")
        else:
            sync_result = self._sync_local_with_s3_objects(
                s3_nodes, s3_keys, source_keys, force_delete=force_delete
            )
            if sync_result.synced:
                logger.debug(
                    "Local filesystem is clean but S3 has changed so we have "
                    "to sync here with S3 so that S3 changes would appear here.",
                )
                logger.info(f"Syncing {self.root.name} with S3...")

        self.config_parser["sync"]["dirty"] = "false"
        if sync_result.synced:
            logger.info("Synced.")
            logger.info(sync_result)
        else:
            logger.info("Everything is up to date.")

    def sync_local(self, force_delete: bool = False) -> SyncResult:
        """Sync the local fs with S3 changes"""
        with self.lock:
            s3_nodes = self._get_node_schema_from_bucket()
            s3_paths = {node.abspath for node in s3_nodes}
            source_paths = self._get_all_files(self.root)

            return self._sync_local_with_s3_objects(
                s3_nodes, s3_paths, source_paths, force_delete
            )

    def create_on_s3(self, path: Path) -> SyncResult:
        """A file has been created in local FS"""
        with self.lock:
            self._create_object_from_here_on_s3(path)
        return SyncResult(created_files=[str(path)])

    def delete_on_s3(self, path: Path) -> SyncResult:
        """A file has been deleted in local FS"""
        with self.lock:
            key = self.service.get_key_from_path(path)
            self.service.delete_objects(key)
        return SyncResult(deleted_files=[str(path)])

    def update_attrs_on_s3(self, path: Path, attrs: FileAttrs):
        mime_type = get_mime_type(path)

        with self.lock:
            key = self.service.get_key_from_path(path)
            self.service.copy_object(key, attrs, mime_type)
        return SyncResult(updated_files=[str(path)])

    def rename_on_s3(self, old_path: Path, new_path: Path) -> SyncResult:
        with self.lock:
            old_key = self.service.get_key_from_path(old_path)
            new_key = self.service.get_key_from_path(new_path)
            self.service.rename_object(old_key, new_key)
        return SyncResult(updated_files=[str(old_path), str(new_path)])

    def update_on_s3(self, path: Path):
        with self.lock:
            self._create_object_from_here_on_s3(path)
        return SyncResult(updated_files=[str(path)])

    def _sync_local_with_s3_objects(
        self,
        s3_nodes: list[NodeSchema],
        s3_paths: set[Path],
        source_paths: set[Path],
        force_delete: bool = False,
    ) -> SyncResult:
        """Update, Delete and Create files in local FS from S3"""
        to_be_updated = s3_paths & s3_paths
        to_update_nodes = list(filter(lambda n: n.abspath in to_be_updated, s3_nodes))

        try:
            with ThreadPoolExecutor(
                self._get_thread_count(to_update_nodes),
                "sync_update",
            ) as executor:
                for node in to_update_nodes:
                    executor.submit(self._update_object_in_local_from_s3, node)
        except ValueError:
            pass

        to_be_created = s3_paths - source_paths
        to_create_nodes = list(filter(lambda n: n.abspath in to_be_created, s3_nodes))

        try:
            with ThreadPoolExecutor(
                self._get_thread_count(to_be_created),
                "sync_create",
            ) as executor:
                for node in to_create_nodes:
                    executor.submit(self._create_object_in_local_from_s3, node)
        except ValueError:
            pass

        to_be_deleted = source_paths - s3_paths

        if force_delete:
            try:
                with ThreadPoolExecutor(
                    self._get_thread_count(to_be_deleted),
                    "sync_delete",
                ) as executor:
                    for file_path in to_be_deleted:
                        executor.submit(os.remove, file_path)
            except ValueError:
                pass

            return SyncResult(
                list(map(lambda n: str(n.abspath), to_create_nodes)),
                list(map(str, to_be_deleted)),
            )

        if to_be_deleted:
            prompt = Prompt()
            res = prompt.ask(
                f"FS sync mechanism is going to delete: {list(map(str, to_be_deleted))} from local FS"
                f"\n'Should it precede?'",
                choices=["yes", "NO"],
                default="NO",
            )
            if res == "yes":
                with ThreadPoolExecutor(
                    self._get_thread_count(to_be_deleted),
                    "sync_delete",
                ) as executor:
                    for file_path in to_be_deleted:
                        executor.submit(os.remove, file_path)
            else:
                to_be_deleted = []
        return SyncResult(
            list(map(lambda n: str(n.abspath), to_create_nodes)),
            list(map(lambda n: str(n.abspath), to_update_nodes)),
            list(map(str, to_be_deleted)),
        )

    def _sync_s3_with_local_objects(
        self,
        s3_nodes: list[NodeSchema],
        s3_paths: set[Path],
        source_paths: set[Path],
        force_delete: bool = False,
    ) -> SyncResult:
        """Update, Upload and Delete objects of S3 that are new in local FS"""
        to_be_updated = s3_paths & s3_paths
        to_update_nodes = list(filter(lambda n: n.abspath in to_be_updated, s3_nodes))

        try:
            with ThreadPoolExecutor(
                self._get_thread_count(to_update_nodes),
                "sync_update",
            ) as executor:
                for node in to_update_nodes:
                    executor.submit(self._update_object_from_local_on_s3, node)
        except ValueError:
            pass

        to_be_created = source_paths - s3_paths

        try:
            with ThreadPoolExecutor(
                self._get_thread_count(to_be_created),
                "sync_upload",
            ) as executor:
                for file_path in to_be_created:
                    executor.submit(
                        self._create_object_from_here_on_s3,
                        file_path,
                    )
        except ValueError:
            pass

        to_be_deleted = s3_paths - source_paths
        if force_delete:
            to_delete_keys: list[NodeSchema] = list(
                map(
                    lambda n: n.key,
                    filter(lambda n: n.abspath in to_be_deleted, s3_nodes),
                )
            )
            self.service.delete_objects(*to_delete_keys)
            return SyncResult(
                list(map(str, to_be_created)),
                list(map(lambda n: str(n.abspath), to_update_nodes)),
                list(map(str, to_be_deleted)),
            )

        if to_be_deleted:
            prompt = Prompt()
            res = prompt.ask(
                f"FS sync mechanism is going to delete: {list(map(str, to_be_deleted))} from S3"
                f"\n'Should it precede?'",
                choices=["yes", "NO"],
                default="NO",
            )
            if res == "yes":
                to_delete_keys: list[NodeSchema] = list(
                    map(
                        lambda n: n.key,
                        filter(lambda n: n.abspath in to_be_deleted, s3_nodes),
                    )
                )
                self.service.delete_objects(*to_delete_keys)
            else:
                to_be_deleted = []
        return SyncResult(
            list(map(str, to_be_created)),
            list(map(lambda n: str(n.abspath), to_update_nodes)),
            list(map(str, to_be_deleted)),
        )

    def _create_object_from_here_on_s3(self, file_path: Path) -> None:
        """Upload single object to S3"""
        try:
            stat = os.lstat(file_path)
        except FileNotFoundError:
            return

        mime_type = get_mime_type(file_path)
        with file_path.open("br") as f:
            self.service.upload_fileobj(
                f,
                self.service.get_key_from_path(file_path),
                FileAttrs.from_stat_result(stat),
                mime_type,
            )

    def _create_object_in_local_from_s3(self, node: NodeSchema) -> None:
        """Download and create object from S3"""
        node.abspath.parents[0].mkdir(parents=True, exist_ok=True)
        with node.abspath.open("bw") as f:
            self.service.download_fileobj(f, node.key)

    def _update_object_in_local_from_s3(self, node: NodeSchema) -> None:
        path = node.abspath
        st = os.lstat(path)
        if st.st_size != node.size:
            with path.open("bw") as f:
                self.service.download_fileobj(f, node.key)

        if (attrs := node.file_attrs) is not None:
            os.utime(path, (attrs.st_atime, attrs.st_mtime))
            os.chown(path, attrs.st_uid, attrs.st_gid)
            os.chmod(path, attrs.st_mode)

    def _update_object_from_local_on_s3(self, node: NodeSchema) -> None:
        path = node.abspath
        key = self.service.get_key_from_path(path)

        st = os.lstat(path)
        attrs = FileAttrs.from_stat_result(st)

        mime_type = get_mime_type(path)

        if st.st_size != node.size:
            with path.open("br") as f:
                self.service.upload_fileobj(f, key, attrs, mime_type)
        else:
            mime_type = get_mime_type(path)
            self.service.copy_object(key, attrs, mime_type)

    def _get_all_files(self, source: Path) -> set[Path]:
        source_keys = set()
        for df in source.iterdir():
            if df.is_file():
                source_keys.add(df)
            else:
                source_keys |= self._get_all_files(df)
        return source_keys

    def _get_node_schema_from_bucket(self) -> list[NodeSchema]:
        response = self.service.list_objects()
        try:
            return [
                NodeSchema(
                    key=cont.Key,
                    last_modified=cont.LastModified,
                    size=cont.Size,
                    abspath=self.root / cont.Key,
                )
                for cont in response.Contents
            ]
        except TypeError:
            return []

    @staticmethod
    def _get_thread_count(keys: Sized) -> int:
        return min(32, len(keys))



from .cli import parser
from .core.log_config import file_queue_listener

if __name__ == "__main__":
    file_queue_listener.start()

    try:
        args = parser.parse_args()
        try:
            args.cls(args)
        except AttributeError:
            parser.print_help()
            exit(-1)
    finally:
        file_queue_listener.start()



from argparse import ArgumentParser

from .mount_cmd import MountHandler
from .config_cmd import ConfigHandler

parser = ArgumentParser(
    "PyS3FUSE", description="A FUSE layer to mount S3 buckets in your local filesystem."
)
subparsers = parser.add_subparsers(
    title="commands", description="Different commands on PyS3FUSE"
)
config_parser = subparsers.add_parser(
    "config", help="Config different parts of PyS3FUSE"
)
config_parser.add_argument(
    "--show",
    help="Print the current config",
    required=False,
    action="store_true",
)
config_parser.add_argument(
    "-ll",
    "--log-level",
    help="Default log level of PyS3FUSE",
    choices=["debug", "info"],
    required=False,
)
config_parser.add_argument(
    "-mt",
    "--multipart-threshold",
    help="The transfer size threshold for which multipart uploads, downloads, and copies will automatically be triggered.",
    choices=["100KB", "250KB", "500KB", "1MB", "2MB", "5MB", "8MB", "10MB"],
    required=False,
    dest="multipart_threshold",
)
config_parser.add_argument(
    "-xc",
    "--max-concurrency",
    help="The maximum number of threads that will be making requests to perform a transfer.",
    choices=["2", "5", "10", "20"],
    required=False,
    dest="max_concurrency",
)
config_parser.add_argument(
    "-mc",
    "--multipart-chunksize",
    help="The partition size of each part for a multipart transfer.",
    choices=["100KB", "250KB", "500KB", "1MB", "2MB", "5MB", "8MB", "10MB"],
    required=False,
    dest="multipart_chunksize",
)
config_parser.set_defaults(cls=ConfigHandler(config_parser))

mount_parser = subparsers.add_parser("mount", help="Mound an S3 bucket to work with")
mount_parser.add_argument(
    "source",
    type=str,
    help="Directory tree to mirror",
)
mount_parser.add_argument(
    "mountpoint",
    type=str,
    help="Where to mount the file system",
)
mount_parser.add_argument(
    "--bucket", required=True, type=str, help="S3 Bucket you want to mound."
)
mount_parser.add_argument(
    "--debug",
    action="store_true",
    default=False,
    help="Enable debugging output",
)
mount_parser.add_argument(
    "--offline",
    action="store_true",
    default=False,
    help="Set when you don't have connection to your S3 service and want to manipulate the mountpoint on yourself. Later PyS3FUSE will sync your mountpoint with your S3 bucket.",
)
mount_parser.add_argument(
    "--multi-user",
    action="store_true",
    default=False,
    help="If you are online and somebody else has mounted the bucket OR you are online but somebody else is connected to S3 pass this option so that on some syscalls FS will sync it self with S3. Either case S3 has precedence.",
    dest="multi_user",
)
mount_parser.add_argument(
    "-f",
    "--force",
    action="store_true",
    default=False,
    help="In the initial synchronization of program, there maybe some objects that have to be deleted in local FS or S3. Default behaviour is to ask the user whether to delete them or not but if this options is passed no prompt will be shown and all will be deleted.",
    dest="force",
)
mount_parser.set_defaults(cls=MountHandler())
